Require Import Coq.Bool.Bool.
Require Import Coq.Arith.Arith.
Require Import Coq.Lists.List.
Require Import Coq.Program.Equality.
Require Import Coq.omega.Omega.

Require Export labelmap.
Require Export lattice.

Module CoreLanguage.

Import ListNotations.

Import LabelMap.
Import SecurityLattice.

(* id for variables*)
Inductive id : Type :=
  | Id : nat -> id
  | ARG.

Hint Constructors id.

Definition beq_id id1 id2 :=
  match id1,id2 with
    | Id n1, Id n2 => beq_nat n1 n2
    | ARG, ARG => true
    | _,_ => false
  end.

Inductive tp : Type :=
  | Tbool
  | Tnat
  | Tref (t : tp).

Definition loc := nat.

Inductive val : Type :=
  | v_bool (b:bool)
  | v_nat (n: nat)
  | v_loc (l: loc).


Inductive exp : Type :=
  | e_var (x: id) : exp
  | e_val (v:val) : exp
  | e_eq : exp->exp->exp
  | e_deref : exp -> exp.

Definition e_bool (b: bool) : exp := e_val (v_bool b).
Definition e_nat (n: nat) : exp := e_val (v_nat n).
Definition e_loc (l: loc) : exp := e_val (v_loc l).

Definition mem : Type := list val.
Definition empty_mem : mem := nil.

Fixpoint update_mem (m:mem) (l:loc) (v:val) : mem := 
  match m with 
  | nil => nil
  | h :: t => match l with
      | O => v :: t
      | S n => h :: (update_mem t n v)
      end
  end.

Definition lookup_mem (m:mem) (l:loc) : val :=
  nth l m (v_bool false).

Inductive value : exp -> Prop :=
  | Value v : value (e_val v).


Inductive cmd : Type :=
  | Raise (l:label)
  | Decl (l:label)
  | Call (ai : aid) (e : exp) (expect_return: bool)
  | Return (e : exp)
  | Exit (e : exp).

Inductive stm : Type :=
  | s_seq (s1 s2 : stm)
  | s_if (e:exp) (s1 s2:stm)
  | s_null
  | s_assign (e1 e2: exp)
  | s_let  (x:id) (e: exp) (s:stm)
  | s_new (x:id) (e: exp) (s:stm)
  | s_cmd (c : cmd).

Fixpoint subst_exp  (x:id) (v:val) (e:exp) : exp :=
  match e with
  | e_var y => if beq_id x y then (e_val v) else e
  | e_eq e1 e2 => e_eq (subst_exp x v e1) (subst_exp x v e2)
  | e_val v => e
  | e_deref e => e_deref (subst_exp x v e)
  end.

Definition subst_cmd (x:id) (v:val) (c:cmd) : cmd :=
  match c with
  | Call ai e er => Call ai (subst_exp x v e) er
  | Return e => Return (subst_exp x v e)
  | Exit e => Exit (subst_exp x v e)
  | Raise _ => c
  | Decl _ => c
  end.


Fixpoint subst_stm (x:id) (v:val) (s:stm) : stm :=
  match s with
  | s_seq s1 s2 => s_seq (subst_stm x v s1) (subst_stm x v s2)
  | s_if e s1 s2 => s_if (subst_exp x v e) (subst_stm x v s1) (subst_stm x v s2)
  | s_null => s_null
  | s_assign e1 e2 => s_assign (subst_exp x v e1) (subst_exp x v e2)
  | s_let y e s =>
    if beq_id x y then s_let x (subst_exp x v e)  s else s_let y (subst_exp x v e) (subst_stm x v s)
  | s_new y e s =>
    if beq_id x y then s_new x (subst_exp x v e)  s else s_new y (subst_exp x v e) (subst_stm x v s)
  | s_cmd c => s_cmd (subst_cmd x v c)
  end.

Inductive step_exp : mem -> exp -> exp -> Prop :=
  | se_eq1 m b1 b2 :  step_exp m (e_eq (e_bool b1) (e_bool b2) ) (e_bool (eqb b1 b2))
  | se_eq2 m n1 n2 :  step_exp m (e_eq (e_nat n1) (e_nat n2) ) (e_bool (beq_nat n1 n2))
  | se_eq3 m l1 l2 :  step_exp m (e_eq (e_loc l1) (e_loc l2) ) (e_bool (beq_nat l1 l2))
  | se_eq4 m v e2 e2' : step_exp m e2 e2' ->  step_exp m (e_eq (e_val v) e2 ) (e_eq (e_val v) e2' )
  | se_eq5 m  e1 e1' e2 : step_exp m e1 e1' ->  step_exp m (e_eq e1 e2 ) (e_eq e1' e2 )
  | se_deref1 m e e': step_exp m e e' -> step_exp m (e_deref e) (e_deref e')
  | se_deref2 m l v : lookup_mem m l = v -> step_exp m (e_deref (e_loc l)) (e_val v).

Inductive step_cmd : mem -> cmd -> cmd -> Prop :=
  | sc_call m e e' ai er: step_exp m e e' -> step_cmd m (Call ai e er) (Call ai e' er)
  | sc_return m e e': step_exp m e e' -> step_cmd m (Return e) (Return e')
  | sc_exit m e e': step_exp m e e' -> step_cmd m (Exit e) (Exit e').

Inductive step_stm : mem -> stm -> mem -> stm -> Prop :=
  | ss_seq1 m s: step_stm m (s_seq s_null s) m s
  | ss_seq2 m m' s1 s1' s2: step_stm m s1 m' s1' -> step_stm m (s_seq s1 s2) m' (s_seq s1' s2)
  | ss_if1 s_true s_false m :
      step_stm m (s_if (e_bool true) s_true  s_false) m s_true
  | ss_if2 s_true s_false m :
      step_stm m (s_if (e_bool false) s_true  s_false) m s_false
  | ss_if3 s_true s_false m e e'  :
      step_exp m e e' -> step_stm m (s_if e s_true s_false)  m (s_if e' s_true s_false)
  | ss_assign1 l m v: step_stm m (s_assign (e_loc l) (e_val v)) (update_mem m l v) s_null
  | ss_assign2 m e e1 e': step_exp m e e' -> step_stm m (s_assign e e1) m (s_assign e' e1)
  | ss_assign3 m l e e': step_exp m e e' -> step_stm m (s_assign (e_loc l) e) m (s_assign (e_loc l) e')
  | ss_let1 m x v s : step_stm m (s_let x (e_val v) s) m (subst_stm x v s)
  | ss_let2 m x e e' s : step_exp m e e' -> step_stm m (s_let x e s) m (s_let x e' s)
  | ss_new1 m x v s:  step_stm m ( s_new x (e_val v) s) (m ++ [v]) (subst_stm x (v_loc (length m)) s)
  | ss_new2 m x e e' s : step_exp m e e' -> step_stm m (s_new x e s) m (s_new x e' s)
  | ss_cmd m cm cm' : step_cmd m cm cm' -> step_stm m (s_cmd cm) m (s_cmd cm').

Definition cntx := id ->  option tp.
Definition empty_cntx : cntx := fun _ => None.

Definition update_cntx : cntx -> id -> tp -> cntx := fun c x t =>
  fun y => if beq_id x y then Some t else c y.

Definition default_cntx : cntx := update_cntx empty_cntx ARG Tnat.


Definition mem_tp := list tp.
Definition lookup_mem_tp (mt: mem_tp) (l: loc) : tp :=
  nth l mt Tbool.

Inductive extends : mem_tp -> mem_tp -> Prop :=
  | extends_nil mt' : extends mt' nil
  | extends_cons v mt' mt : extends mt' mt -> extends (v::mt') (v::mt).

Inductive wt_exp : cntx -> mem_tp -> exp -> tp -> Prop :=
  | wte_var c mt x t:  (c x) = Some t -> wt_exp c mt (e_var x) t
  | wte_bool c mt b: wt_exp c mt (e_bool b) Tbool
  | wte_nat c mt n: wt_exp c mt (e_nat n) Tnat
  | wte_loc c mt l: l < length mt -> wt_exp c mt (e_loc l) (Tref (lookup_mem_tp mt l))
  | wte_eq c mt e1 e2 t: wt_exp c mt e1 t -> wt_exp c mt e2 t -> wt_exp c mt (e_eq e1 e2) Tbool
  | wte_deref c mt e t: wt_exp c mt e (Tref t) -> wt_exp c mt (e_deref e) t.

Inductive wt_cmd : cntx -> mem_tp -> cmd -> Prop :=
  | wtc_raise c mt l: wt_cmd c mt (Raise l)
  | wtc_decl c mt l: wt_cmd c mt (Decl l)
  | wtc_call c mt ai e er: wt_exp c mt e Tnat -> wt_cmd c mt (Call ai e er)
  | wtc_return c mt e: wt_exp c mt e Tnat -> wt_cmd c mt (Return e)
  | wtc_exit c mt e: wt_exp c mt e Tnat -> wt_cmd c mt (Exit e).

Hint Constructors wt_cmd.

Inductive wt_stm : cntx -> mem_tp -> stm -> Prop :=
  | wts_seq c mt s1 s2 : wt_stm c mt s1 -> wt_stm c mt s2 -> wt_stm c mt (s_seq s1 s2)
  | wts_if c mt e s1 s2 : wt_exp c mt e Tbool -> wt_stm c mt s1 -> wt_stm c mt s2 ->
    wt_stm c mt (s_if e s1 s2)
  | wts_assign c mt e1 e2 t: wt_exp c mt e1 (Tref t) -> wt_exp c mt e2 t -> wt_stm c mt (s_assign e1 e2)
  | wts_let c mt x e s t: wt_exp c mt e t -> wt_stm (update_cntx c x t) mt s -> wt_stm c mt (s_let x e s)
  | wts_new c mt x e s t: wt_exp c mt e t -> wt_stm (update_cntx c x (Tref t)) mt s -> wt_stm c mt (s_new x e s)
  | wts_null c mt : wt_stm c mt s_null
  | wts_cmd c mt cm: wt_cmd c mt cm -> wt_stm c mt (s_cmd cm).

Definition wt_mem (m:mem) (mt:mem_tp) : Prop :=
  length m = length mt /\ 
  (forall (l:loc), wt_exp empty_cntx mt (e_val (lookup_mem m l)) (lookup_mem_tp mt l)).

Inductive free_in_exp : id -> exp -> Prop :=
  | fie_var x : free_in_exp x (e_var x)
  | fie_eq1 x e1 e2 : free_in_exp x e1 -> free_in_exp x (e_eq e1 e2)
  | fie_eq2 x e1 e2 : free_in_exp x e2 -> free_in_exp x (e_eq e1 e2)
  | fie_deref x e : free_in_exp x e -> free_in_exp x (e_deref e).

Inductive free_in_cmd : id -> cmd -> Prop :=
  | fic_call x e ai er: free_in_exp x e -> free_in_cmd x (Call ai e er)
  | fic_return x e: free_in_exp x e -> free_in_cmd x (Return e)
  | fic_exit x e: free_in_exp x e -> free_in_cmd x (Exit e).

Inductive free_in_stm : id -> stm -> Prop :=
  | fis_seq1 x s1 s2 : free_in_stm x s1 -> free_in_stm x (s_seq s1 s2)
  | fis_seq2 x s1 s2 : free_in_stm x s2 -> free_in_stm x (s_seq s1 s2)
  | fis_ifc x s1 s2 e : free_in_exp x e -> free_in_stm x (s_if e s1 s2)
  | fis_if1 x s1 s2 e : free_in_stm x s1 -> free_in_stm x (s_if e s1 s2)
  | fis_if2 x s1 s2 e : free_in_stm x s2 -> free_in_stm x (s_if e s1 s2)
  | fis_assign1 x e1 e2 : free_in_exp x e1 -> free_in_stm x (s_assign e1 e2)
  | fis_assign2 x e1 e2 : free_in_exp x e2 -> free_in_stm x (s_assign e1 e2)
  | fis_lete x y e s :  free_in_exp x e -> free_in_stm x (s_let y e s)
  | fis_lets x y e s : x <> y -> free_in_stm x s -> free_in_stm x (s_let y e s)
  | fis_newe x y e s :  free_in_exp x e -> free_in_stm x (s_new y e s)
  | fis_news x y e s : x <> y -> free_in_stm x s -> free_in_stm x (s_new y e s)
  | fis_cmd x cm: free_in_cmd x cm -> free_in_stm x (s_cmd cm).

End CoreLanguage.
