Require Import Coq.Bool.Bool.
Require Import Coq.Arith.Arith.
Require Import Coq.Lists.List.
Require Import Coq.Program.Equality.
Require Import Omega.
Require Import Coq.Program.Tactics.
Require Export Coq.Structures.Equalities.

Require Export labelmap.
Require Export lattice.
Require Export core.
Require Export system.
Require Export sys_preservation.

Module Projection.


Import ListNotations.

Import LabelMap.
Import SecurityLattice.
Import CoreLanguage.
Import System.
Import Preservation.


Fixpoint proj_dll (n: nat) (k : label) (dll : list (iid * label) ) : list (iid * label) :=
  match dll with
  | [] => []
  | (ii, l) :: dll => if lte_bool l k then
      (Iid n , l) :: proj_dll (S n) k dll else proj_dll n k dll
  end.

Fixpoint proj_ii_dll (n: nat) (k: label) (dll : list (iid * label)) (ii:iid) : option iid :=
match dll with
| [] => None
| (ii2,l) :: dll =>
  if beq_iid ii ii2 then
    if lte_bool l k then Some (Iid n) else None
  else
    if lte_bool l k then proj_ii_dll (S n) k dll ii else proj_ii_dll n k dll ii 
end.

Definition proj_ii_alm (k: label) (alm : option app_label_map)  (ii:iid) : option iid :=
  match alm with
  | Some (Single_Instance_App_Launched sl dl ii2) =>
      if (beq_iid ii2 ii)  && (lte_bool dl k) then Some ii else None
  | Some (Multi_Instance_App _ dll) => proj_ii_dll 0 k dll ii
  | _ => None
  end.

Definition proj_ii_lm (k: label) (lm: label_map) (ai: aid) (ii:iid) : option (aid*iid) :=
  match proj_ii_alm k (get_app_label_map lm ai) ii with
  | None => None
  | Some ii' => Some (ai,ii')
  end.

Definition proj_ra (k: label) (lm: label_map) (ra : ret_addr) : ret_addr :=
  match ra with
  | None => None
  | Some (ai,ii) => proj_ii_lm k lm ai ii
  end.

Fixpoint proj_ais_m (n: nat) (k: label) (lm: label_map) (dll : list (iid * label)) ais : app_inst_set :=
  match dll, ais with
  | (_,l) :: dll, Inst _ m s ra :: ais =>
      if lte_bool l k then
        Inst (Iid n) m s (proj_ra k lm ra) :: proj_ais_m (S n) k lm dll ais
      else proj_ais_m n k lm dll ais
  | _ , _ => []
  end.


Definition proj_alm (k : label) (alm : app_label_map) : app_label_map:=
  match alm with
  | Multi_Instance_App sl dll => Multi_Instance_App sl (proj_dll 0 k dll)
  | Single_Instance_App_Unlaunched sl prev => if lte_bool sl k then
                         alm else NA
  | Single_Instance_App_Launched sl dl ii => if lte_bool dl k then
                         alm else NA
  | NA => NA
  end.


Definition proj_ra_inst (k : label) (lm : label_map) (inst : instance) : instance :=
  match inst with
  | Inst ii m s ra => Inst ii m s (proj_ra k lm ra) end.

Definition proj_ais (k : label) (lm : label_map) (alm : app_label_map) ais : app_inst_set:=
  match alm with
  | Single_Instance_App_Unlaunched sl prev => if lte_bool sl k then
                         map (proj_ra_inst k lm) ais else []
  | Single_Instance_App_Launched sl dl ii => if lte_bool dl k then
                         map (proj_ra_inst k lm) ais else []
  | Multi_Instance_App sl dll => (proj_ais_m 0 k lm dll ais)
  | NA => []
  end.


Definition proj_lm (k : label) (lm: label_map) : label_map :=
  map (proj_alm k) lm.

Definition proj_is (k: label) (LM: label_map) (lm: label_map) (is : inst_set) : inst_set :=
  map (fun a => proj_ais k LM (fst a) (snd a)) (combine lm is).

Definition proj_pend (k: label) (LM: label_map)
    (alm : option app_label_map) (pt: pendingType) : option pendingType :=
  match pt with
  | pCall ra => match alm with 
    | Some (Single_Instance_App_Unlaunched l _)
    | Some (Single_Instance_App_Launched _ l _)
    | Some (Multi_Instance_App l _) =>
        if lte_bool l k then Some (pCall (proj_ra k LM ra) ) else None
    | _ => None
    end
  | pReturn ii => match alm with
    | Some (Single_Instance_App_Launched _ dl ii2) =>
        if beq_iid ii ii2 && lte_bool dl k then Some pt else None 
    | Some (Multi_Instance_App _ dll) =>
        match proj_ii_dll 0 k dll ii with
      | Some ii' => Some (pReturn ii') | None => None end
    | _ => None
    end
  end.

Definition proj_pending (k: label) (LM: label_map) (lm: label_map) (p: pending) : option pending :=
  match p with 
  | Pend callee n dl pt =>
    match proj_pend k LM (get_app_label_map lm callee) pt with
    | None => None
    | Some pt => Some (Pend callee n dl pt)
    end
  end.

Fixpoint proj_pq (k : label) (LM:label_map) (lm: label_map) (pq: pqueue) : pqueue :=
  match pq with
  | [] => []
  | p :: pq =>
    match proj_pending k LM lm p with
    | None => proj_pq k LM lm pq
    | Some p => p :: proj_pq k LM lm pq
    end
  end.


Definition proj (k : label) (s: sys) : sys :=
  match s with (Sys apps lm is pq) =>
      (Sys apps (proj_lm k lm) (proj_is k lm lm is) (proj_pq k lm lm pq))
  end.
  
Definition proj_trace (k : label) (t: trace) : trace :=
  filter (fun a => lte_bool (snd a) k) t.

End Projection.
